/* Class for drawing and animating "Firefly"-dots inside google maps.
*
* MovementOverlay is the main class, which will keep track of the firefly-objects'
* size and movement. Firefly-objects and the animation are controlled through this class.
*
* MovementOverlay requires google maps object when initialized. Google Maps API V3 was used
* for development, so using older/newer versions with this might not work. The animation will be
* drawn atop of this map. Overlay then needs firefly-objects to animate, which can be added 
* one by one with add_object()-method. Animation can then be started and stopped by calling 
* MovementOverlay's toggle_animation()-method with bool value as a paramater. Method 
* check_animation() can be used to check if there is any object left to animate and method
* reset_animation() to reset animation's state as it was in the beginning. Method delete_overlay()
* will delete every created firefly-object, but the MovementOverlay-object is not destroyed.
*
*
* Author: Henri Timonen
*/

/*
* Constructor for firefly overlay.
* Takes the google maps object as a parameter, 
* to where the firefly-objects will be drawn.
*/
function MovementOverlay(map) {
	this.objectArray = [];
	this.movementArray = [];
	this.visibility = true;
	this.map = map;
	this.firefly_animation = false;
}

/* Creates a firefly-object and pushes it to the array
*
* properties - an object containing starting coornidates, starting size, ending coordinates, 
*	ending size and object's color. Coordinates are in google.maps.LatLng format and sizes in pixels.
*/
MovementOverlay.prototype.add_object = function(properties) {
	var obj = new firefly_object(properties['startCoord'],
						 properties['endCoord'],
						 properties['startSize'],
						 properties['endSize'],
						 properties['colour']);
						 
	this.objectArray.push(obj);
	this.movementArray.push(obj);
}

//Goes through the array of objects and recalculates the current location and size
//timing tells how long it has been since the last animation frame
MovementOverlay.prototype.move_objects = function(timing) {
	
	var len = this.movementArray.length;
	
	if (len == 0) {
		this.firefly_animation = false;
		return
	}
	var pointer = 0, temp = 0;
	for (var i = 0; i < len; i++) {

		var elem = this.movementArray[pointer];
		temp = 0;
		
		//Set circle's new center point
		if (elem.deltaLat) {
			
			elem.currentCoord = new google.maps.LatLng(elem.currentCoord.lat()+(elem.deltaLat),
													   elem.currentCoord.lng()+(elem.deltaLng));
			elem.mov += 1;
			elem.circle.setCenter(elem.currentCoord);
		}
		
		//Set circle's new size
		if (elem.deltaSize > 0) {
			elem.size -= elem.deltaSize;
			elem.circle.setRadius(elem.size);
		}
		
		//Check if the circle can be removed from the array
		if (!(elem.mov < elem.frames) && !(elem.size > elem.endSize)) {
			this.movementArray.splice(pointer,1);
		}else{
			pointer += 1;
		}
	}
}

//Checks if the animation has already ended and returns bool value
MovementOverlay.prototype.check_animation = function() {
	if (this.movementArray.length == 0) {
		return false;
	}else{
		return true;
	}
}

//Resets the animation
MovementOverlay.prototype.reset_animation = function() {
	this.firefly_animation = false;
	this.movementArray = [];
	for (obj in this.objectArray) {
		var node = this.objectArray[obj];
		node.currentCoord = node.startCoord;
		node.size = node.startSize;
		node.mov = 1;
		node.circle.setCenter(node.currentCoord);
		node.circle.setRadius(node.size);
		this.movementArray.push(node);
	}
}

//Changes the state of the firefly objects' animation, i.e. to on or off.
MovementOverlay.prototype.toggle_animation = function(state) {
	if (state) {
		this.firefly_animation = true;
		startTime = window.mozAnimationStartTime || Date.now();
		this.update();
	}else{
		this.firefly_animation = false;
	}
}

//Requestanimationframe declaration with setTimeout fallback
window.requestAnimFrame = (function() {
	return 	window.requestAnimationFrame ||
			window.mozRequestAnimationFrame || 
			window.webkitRequestAnimationFrame ||
			window.oRequestAnimationFrame ||
			window.msRequestAnimationFrame ||
			function(callback, element) {
				window.setTimeout(callback, 1000/30);
			};
})();
var startTime = 0;

//Updates the animation, calculates the time passed and calls itself again
//timestamp is given only in Firefox, for other browsers Date.now() fallback option
MovementOverlay.prototype.update = function(timestamp) {

	/*
	* Diff tells how long it took to update firefly-objects location and come back.
	* However, it is not used in this current version. To take it into account when
	* updating locations would have required some restructuring while initializing
	* time-to-live values for each firefly-object. Due to time constraints it was left
	* out, and it was not certain if it would have provided much difference without 
	* hurting animation's quality too much.
	*/
	//var diff = (timestamp || Date.now()) - startTime;
	//startTime = (timestamp || Date.now());
	//this.move_objects(diff);
	
	this.move_objects();

	if (this.firefly_animation) requestAnimFrame(this.update.bind(this));
}

//Toggles every firefly objects' visibility according the their previous value.
MovementOverlay.prototype.toggle = function() {
	this.visibility = (this.visibility) ? false : true;
	for (obj in this.objectArray) {
		this.objectArray[obj].circle.setVisible(this.visibility);
	}
}

//Just hides every firefly object.
MovementOverlay.prototype.hide = function() {
	for (obj in this.objectArray) {
		this.objectArray[obj].circle.setVisible(false);
	}
}

//Deletes the whole overlay and every firefly-object.
MovementOverlay.prototype.delete_overlay = function() {
	for (obj in this.objectArray) {
		this.objectArray[obj].circle.setMap(null);
	}
	for (obj in this.movementArray) {
		this.movementArray[obj].circle.setMap(null);
	}
	this.objectArray = [];
	this.movementArray = [];
}


/*
* Single firefly object.
* Coordinates are in google.maps.LatLng format
* Size values are in pixels
* Stationary objects can be created by passing null-value as an endCoord
*/
function firefly_object(startCoord, endCoord, startSize, endSize, colour) {
	this.startCoord = startCoord;
	this.endCoord = endCoord;
	this.currentCoord = startCoord;
	
	this.colour = colour
	
	this.startSize = startSize;
	this.size = startSize;
	this.endSize = endSize;
	
	this.frames = 0;
	this.deltaSize = 0;
	this.distance = 0;
	
	//Keeps track of the movement
	this.mov = 1;
	
	if (this.endCoord) {
		//Set frames according to the distance moved
		var meters = this.calculate_distance(this.startCoord, this.endCoord);
		this.distance = meters;
		
		this.frames = Math.round(Math.pow(meters,0.6));

		this.deltaSize = (startSize - endSize) / this.frames;
		
		this.deltaLat = (endCoord.lat() - startCoord.lat()) / this.frames;
		this.deltaLng = (endCoord.lng() - startCoord.lng()) / this.frames;
	}else{
		this.deltaLat = null;
		this.deltaLng = null;
		
		this.deltaSize = (startSize - endSize) / 50;
	}
	
	this.circleOpt = {
		fillColor: this.colour,
		fillOpacity: 0.5,
		strokeColor: this.colour,
		strokeOpacity: 0.15,
		strokeWeight: 7,
		map: map,
		clickable: false,
		center: this.startCoord,
		radius: this.startSize
	};
	this.circle = new google.maps.Circle(this.circleOpt);
}

//Converts numeric degrees to radians
if (typeof(Number.prototype.toRad) === "undefined"){
	Number.prototype.toRad = function() {
		return this * Math.PI / 180;
	}
}
			
/*Calculates distance between two coordinate points
*
* Parameters:
*	coord1 - starting coordinate
*	coord2 - ending coordinate
*/
firefly_object.prototype.calculate_distance = function(coord1, coord2) {
	var dLat = (coord2.lat()-coord1.lat()).toRad();
	var dLng = (coord2.lng()-coord1.lng()).toRad();
	var lat1 = coord1.lat().toRad();
	var lat2 = coord2.lat().toRad();
				
	var a = Math.sin(dLat/2) * Math.sin(dLat/2) + 
		Math.sin(dLng/2) * Math.sin(dLng/2) * Math.cos(lat1) * Math.cos(lat2);
	var c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a));
	
	return (6371000 * c);
}

//Sets object's size (in pixels)
firefly_object.prototype.set_size = function(size) {
	this.size = size;
}

//Set object's 'time-to-live' (in frames)
firefly_object.prototype.set_frames = function(frames) {
	this.frames = frames;
}